#include <stdio.h>
#include <inttypes.h>

#define SOKOL_DEBUG

#define SOKOL_APP_IMPL
#define SOKOL_GFX_IMPL
#define SOKOL_GLUE_IMPL
#define SOKOL_FETCH_IMPL
#define SOKOL_LOG_IMPL
#define SOKOL_METAL
#include <sokol/sokol_log.h>
#include <sokol/sokol_gfx.h>
#include <sokol/sokol_app.h>
#include <sokol/sokol_glue.h>
#include <sokol/sokol_fetch.h>

#define ARENA_IMPLEMENTATION
#include <tsoding/arena.h>

static Arena default_arena = {0};
static Arena temporary_arena = {0};
static Arena *context_arena = &default_arena;

void *context_alloc(size_t size) {
    // assert(context_arena);
    return arena_alloc(context_arena, size);
}

typedef struct {
    double position[2];
    double size[2];
    double border_size[2];
} GpuUiRect;

typedef struct {
    double atlas_position[2];
    double size[2];
    double position[2];
} GpuGlyph;

typedef struct {
    size_t size;
    size_t capacity;
    uint8_t *data;
} U8Array;

U8Array new_u8array(size_t capacity) {
    return (U8Array) {
        .size = 0,
        .capacity = capacity,
        .data = context_alloc(capacity),
    };
}

void push_u8array(U8Array *array, uint8_t *items, size_t num) {
    if (array->size + num <= array->capacity) {
        memcpy(array->data + array->size, items, num);
        array->size += num;
    } else {
        fprintf(stderr, "failed to push to u8 array, size+num > capacity\n");
    }
}

static struct {
    sg_pass_action pass_action;
    sg_pipeline pip;
    sg_bindings bind;

    sg_image text_atlas_image;
    sg_sampler text_atlas_sampler;

    sg_shader_desc scratch_shader_desc;

    U8Array gpu_ui_rects;
    U8Array gpu_glyphs;

    bool should_exit;
} state;

void vertex_shader_loaded(const sfetch_response_t *response) {
    if (response->fetched) {
        state.scratch_shader_desc.vs.source = response->data.ptr;
        state.scratch_shader_desc.vs.entry = "vs_main";
    } else if (response->failed) {
        fprintf(stderr, "failed to load vertex shader\n");
        exit(1);
    }
}

void fragment_shader_loaded(const sfetch_response_t *response) {
    if (response->fetched) {
        state.scratch_shader_desc.fs = (sg_shader_stage_desc){
            .source = response->data.ptr,
            .entry = "fs_main",
            .images[0].used = true,
            .samplers[0].used = true,
            .image_sampler_pairs[0] = { .used = true, .image_slot = 0, .sampler_slot = 0 },
        };
        state.scratch_shader_desc.fs.source = response->data.ptr;
        state.scratch_shader_desc.fs.entry = "fs_main";
    } else if (response->failed) {
        fprintf(stderr, "failed to load vertex shader\n");
        exit(1);
    }
}

void ed_init() {
    sg_setup(&(sg_desc) {
        .environment = sglue_environment(),
    });

    float vertices[] = {
        // positions            colors
        -0.5f,  0.5f, 0.5f,     0.0f, 0.0f,
         0.5f,  0.5f, 0.5f,     1.0f, 0.0f,
         0.5f, -0.5f, 0.5f,     1.0f, 1.0f,
        -0.5f, -0.5f, 0.5f,     0.0f, 1.0f,
    };
    state.bind.vertex_buffers[0] = sg_make_buffer(&(sg_buffer_desc) {
        .data = SG_RANGE(vertices)
    });

    const uint16_t indices[] = { 0, 1, 2,  0, 2, 3 };
    state.bind.index_buffer = sg_make_buffer(&(sg_buffer_desc) {
        .type = SG_BUFFERTYPE_INDEXBUFFER,
        .data = SG_RANGE(indices)
    });

    sfetch_setup(&(sfetch_desc_t){ .logger.func = slog_func });

    char vs_source[8000] = { 0 };
    char fs_source[8000] = { 0 };

    sfetch_handle_t vs_handle = sfetch_send(&(sfetch_request_t) {
        .path = "./bin/transpiled_shaders/vertex.metal",
        .callback = vertex_shader_loaded,
        .buffer = { .ptr = vs_source, .size = sizeof(vs_source) },
    });
    sfetch_handle_t fs_handle = sfetch_send(&(sfetch_request_t) {
        .path = "./bin/transpiled_shaders/fragment.metal",
        .callback = fragment_shader_loaded,
        .buffer = { .ptr = fs_source, .size = sizeof(fs_source) },
    });

    // block until files are loaded
    while (sfetch_handle_valid(vs_handle) || sfetch_handle_valid(fs_handle)) {
        sfetch_dowork();
    }

    uint8_t *font_bitmap = context_alloc(256*256 * sizeof(uint8_t));
    font_bitmap[4 + 4 * 256] = 255;

    state.text_atlas_sampler = sg_make_sampler(&(sg_sampler_desc) { .mag_filter = SG_FILTER_LINEAR });
    state.text_atlas_image = sg_make_image(&(sg_image_desc) {
        .width = 256,
        .height = 256,
        .pixel_format = SG_PIXELFORMAT_R8,
        .data.subimage[0][0] = { .ptr = font_bitmap, .size = 256*256 * sizeof(uint8_t) },
    });
    state.bind.fs.images[0] = state.text_atlas_image;
    state.bind.fs.samplers[0] = state.text_atlas_sampler;

    sg_shader shd = sg_make_shader(&state.scratch_shader_desc);
    state.pip = sg_make_pipeline(&(sg_pipeline_desc) {
        .shader = shd,
        .index_type = SG_INDEXTYPE_UINT16,
        .layout = {
            .attrs = {
                [0] = { .offset=0, .format=SG_VERTEXFORMAT_FLOAT3 },
                [1] = { .offset=12, .format=SG_VERTEXFORMAT_FLOAT2 },
            },
        },
    });

    state.gpu_ui_rects = new_u8array(sizeof(GpuUiRect) * 2000);
    state.gpu_glyphs = new_u8array(sizeof(GpuGlyph) * 8000);
}
void ed_frame() {
    sg_begin_pass(&(sg_pass) { .action = state.pass_action, .swapchain = sglue_swapchain() });
    {
        sg_apply_pipeline(state.pip);
        sg_apply_bindings(&state.bind);
        sg_draw(0, 6, 1);
    }
    sg_end_pass();
    sg_commit();
}
void ed_cleanup() {
    sfetch_shutdown();
    sg_shutdown();
}
void ed_event(const sapp_event *event) {
    switch (event->type) {
        case SAPP_EVENTTYPE_MOUSE_DOWN:
            if (event->mouse_button == SAPP_MOUSEBUTTON_LEFT) {
                sapp_lock_mouse(true);
            }
            break;

        case SAPP_EVENTTYPE_MOUSE_UP:
            if (event->mouse_button == SAPP_MOUSEBUTTON_LEFT) {
                sapp_lock_mouse(false);
            }
            break;

        default:
            break;
    }
}

sapp_desc sokol_main(int argc, char *argv[]) {
    return (sapp_desc) {
        .width = 640,
        .height = 480,
        .init_cb = ed_init,
        .frame_cb = ed_frame,
        .cleanup_cb = ed_cleanup,
        .event_cb = ed_event,
        .icon.sokol_default = true,
        .logger.func = slog_func,
    };
}
